/*
  Include-file    get_tab_size.inc
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE v4 upwards, all variants
  Version         v2.0.1   18 May 2025


  For macro programmers this include-file implements the function

    integer get_tab_size()

  It returns the current buffer's best tab size, or 0 if none could be
  determined.

  Here tab size means how much space is consistently used to increase lines'
  indentations.

  See the "ALGORITHM" topic in the "TECHNICAL BACKGROUND INFORMATION" section
  if you really want to know some of the dirty details.

  This include-file comes with a tool get_tab_size_test to test and demonstrate
  its use.

  Terminology:
    "TabWidth"
      is the name of a TSE setting used for editing all buffers.
    "Tab size"
      can either refer to an editor setting in general, or to the editor
      setting that apparently was used to edit a specific file.
      In the latter case we can say that the file has a tab size.


  DISCLAIMER

    For a tiny minority of buffers no sensible tab size can be determined,
    for example for empty buffers, or for buffers that use different tab sizes
    in different parts of the buffer, or for buffers that otherwise do not use
    consistent tab sizes, or for buffers that only use tab characters for
    indentation.
    And in an even tinier minority of cases get_tab_size may simply fail to
    determine what the human intended.


  USAGE

  - Either include get_tab_size.inc in your macro with an #include statementment
    like
      #include ["get_tab_size.inc"]
    or copy/paste get_tab_size.inc's source code as a whole into your macro.

    I prefer the copy/paste method. Because I distribute macros, I do not want
    to impose on users of my macros the need to do dependency management of the
    versions of macros and the versions of something as technical as .inc files.

    An great example of this is, that this specific version of get_tab_size.inc
    includes a copy/pasted copy of a specific version of int_array.inc's source
    code.

  - For debugging get_tab_size() you can put
      #define GET_TAB_SIZE_DEBUG 1
    before get_tab_size.inc's code.

    This will create a new buffer with get_tab_size()'s intermediate results.


  HISTORY

  v1      24 Aug 2021
    Initial release.
    Only worked for files without tab characters.


  v2      16 May 2025
    Now also works for files with tab characters.

    Changes in v2 that break backwards compatibility with v1:
    - If no tab size can be determined, then now 0 is returned instead of 1.
      For example for empty files, files with no indentation, or files that
      only use Tab characters for indentation.
    - get_tab_size.inc is now an include file:
      It is no longer executable as a macro, and no longer has parameters.
    - It can no longer create a debug report.


  v2.0.1  18 May 2025
    Fixed "ERROR 6" warning.

*/



/*
  T E C H N I C A L    B A C K G R O U N D   I N F O R M A T I O N


  ALGORITHM

  To determine the current buffer's tab size the tool counts the relative
  increase in indentation between consecutive lines. Tests show that this
  simple method works better than other simple methods.

  Take this example, that only has tabbed indentation.
  The example uses the ^ character to indicate the start of a line:
    ^if a
    ^  if b
    ^    x = 1
    ^  endif
    ^endif
  The 5 lines have absolute indentations 0, 2, 4, 2 and 0.
  The algorithm only looks at relative increases, so between line 1 and 2 and
  between line 2 and 3, giving tab sizes 2 and 2.

  The algorithm is even a bit smarter than that.

  The following example has tabbed and alignment indentation:
    ^if a
    ^  if (   b >= 1
    ^     and b <= 3 )
    ^    x = 1
    ^  endif
    ^endif
  Line 3 does not have tabbed indentation but alignment indentation!
  The 5 lines have absolute indentations 0, 2, 5, 4, 2 and 0 respectively.
  Between lines 2 and 3 the indentation increases with 3 to a total of 5, but
  because 5 is not a multiple of 3, this cannot be tabbed intdentation, and the
  line is ignored.
  The algorithm therefore sees an indentation increase of 2 between lines 2
  and 4, and counts that as a possible tab size of 2. Allrighty!

  Of course there are sequences of lines that make the algorithm fail, but
  tests in practice so far show, that the determined tab size for the whole
  buffer is always defendable, even though in a few worst cases that might be
  because there is no one sensible tab size for a particular buffer.


  TAB SEPARATED VALUES

    Tab-separated value (TSV) files have nothing to to with indentation.

    Like the frequntly occurring comma-separated value files and
    semi-colon-separated value files, in theory tab-separated value files can
    also exist, and I wondered whether to handle them as an exceptional case.

    I scanned my computer, which has software from many different sources, and
    found that TSV files do exist in the wild, but are very rare.

    So rare, that I decided not to handle them.

    Note:
      I found that TSV files usually does not have extension .tsv.

    Note:
      Here is a related, off-putting technical tidbit I encountered.
      If the last field of a line in a TSV file is empty,
      and TSE's RemoveTrailingWhite setting is ON,
      then TSE uses non-intuitive rules for removing trailing tab characters.
      For eample:
        On the first line a trailing tab character is removed when the file is
        loaded.
        A trailing tab character is removed after it is searched and found,
        so you will find it only once.
      This behavior is not necessarily wrong, but might be unexpected.

*/





//  Start of int_array implementation ...

/*
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Version         v1   14 May 2025
  Compatibility   TSE v4 upwards, all variants

  Only use the following functions to manipulate integer arrays:
    integer proc int_array_create(string array_name)
    integer proc int_array_delete(string array_name)
    integer proc int_array_set(string array_name, integer index, integer value)
    integer proc int_array_get(string array_name, integer index)

  int_array_get() returns the value at the index, which defaults to 0.
  All other functions return TRUE or FALSE if they went ok or not.

  An array is intended to only be used in the macro it is created in.
  An array name only needs to be unique within the macro it is created in.
  You can create as many arrays as TSE has free memory and bufferids.
  Creating an array will use up a bufffer id, which limits you to MAXINT
  arrays minus how many bufferids other TSE programs use.
  Array elements can have an index from MININT to MAXINT.
  When an array is created, all array elements exist with value 0.
  Array elements with value 0 do not use memory.
  Setting an array element from not 0 to 0 frees up its memory usage.
  Deleting an array will free up all its memory usage at once.


  NOTE
    I maintain int_array's source as a separate include-file,
    and as an independent and versioned source of truth.

    But for actual use I copy int_array's source as a whole into each macro
    that needs it, so that each such a macro can be tested and distributed
    without int_arrays's version being an outside dependency.

*/


// Do not use or call these int_array_internal_... variables and functions!

integer int_array_internal_id                    = 0
string  int_array_internal_prefix [MAXSTRINGLEN] = 'IntArray'


proc int_array_internal_context()
  integer org_id = 0

  if int_array_internal_id == 0
    int_array_internal_prefix = SplitPath(CurrMacroFilename(), _NAME_) + ':IntArray'
    org_id                    = GetBufferId()
    int_array_internal_id     = CreateTempBuffer()
    ChangeCurrFilename(int_array_internal_prefix, _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
    int_array_internal_prefix = int_array_internal_prefix + ':'
    GotoBufferId(org_id)
  endif
end int_array_internal_context


proc int_array_internal_warn(string array_name, string msg)
  if (Query(MsgLevel) in _ALL_MESSAGES_, _WARNINGS_ONLY_)
    if Query(Beep)
      Alarm()
    endif

    MsgBox(int_array_internal_prefix + array_name,
           iif(Val(msg),
               'ERROR ' + msg + '.',
               'ERROR: ' + msg))
  endif
end int_array_internal_warn

// End of int_array_internal_... variables and functions.


// Start of externally callable int_array functions:

integer proc int_array_create(string array_name)
  integer array_id = 0
  integer ok       = FALSE
  integer org_id   = GetBufferId()

  int_array_internal_context()

  array_id = GetBufferInt(int_array_internal_prefix + array_name,
                          int_array_internal_id)
  if array_id
    int_array_internal_warn(array_name, '1')
  else
    array_id = CreateTempBuffer()

    if array_id
      ChangeCurrFilename(int_array_internal_prefix + array_name,
                         _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
      if SetBufferInt(int_array_internal_prefix + array_name,
                      array_id,
                      int_array_internal_id)
        ok = TRUE
      else
        int_array_internal_warn(array_name, '2')
      endif
    else
      int_array_internal_warn(array_name, '3')
    endif

    GotoBufferId(org_id)
  endif

  return(ok)
end int_array_create


integer proc int_array_delete(string array_name)
  integer array_id = 0
  integer ok       = FALSE

  int_array_internal_context()
  array_id = GetBufferInt(int_array_internal_prefix + array_name,
                          int_array_internal_id)
  if array_id
    ok = TRUE

    if not AbandonFile(array_id)
      int_array_internal_warn(array_name, '4')
      ok = FALSE
    endif

    if not DelBufferVar(int_array_internal_prefix + array_name,
                        int_array_internal_id)
      int_array_internal_warn(array_name, '5')
      ok = FALSE
    endif
  else
    int_array_internal_warn(array_name, '6')
  endif

  return(ok)
end int_array_delete


integer proc int_array_set(string array_name, integer index, integer value)
  integer array_id = 0
  integer ok       = FALSE

  int_array_internal_context()
  array_id = GetBufferInt(int_array_internal_prefix + array_name,
                          int_array_internal_id)

  if array_id
    if value
      if SetBufferInt(int_array_internal_prefix + array_name + Str(index),
                      value,
                      array_id)
        ok = TRUE
      else
        int_array_internal_warn(array_name, '7')
      endif
    else
      DelBufferVar(int_array_internal_prefix + array_name + Str(index),
                   array_id)
    endif
  else
    int_array_internal_warn(array_name, '8')
  endif

  return(ok)
end int_array_set


integer proc int_array_get(string array_name, integer index)
  integer array_id = 0
  integer value    = 0

  int_array_internal_context()
  array_id = GetBufferInt(int_array_internal_prefix + array_name,
                          int_array_internal_id)

  if array_id
    value = GetBufferInt(int_array_internal_prefix + array_name + Str(index),
                         array_id)
  endif

  return(value)
end int_array_get

// End of int_array implementation.





// Get tab size implementation


#ifdef GET_TAB_SIZE_DEBUG
  proc get_tab_size_debug_report(integer best_tab_size,
                                 integer max_tab_size)
    string  curr_file_name [MAXSTRINGLEN] = CurrFilename()
    integer first_log_line                = 0
    integer last_log_line                 = 0
    integer tab_size                      = 0

    NewFile()
    SetUndoOff()
    AddLine('')
    AddLine(curr_file_name)
    AddLine('')
    AddLine(Format('Tab Size':10, 'Occurs':10))
    first_log_line = CurrLine() + 1

    for tab_size = 1 to max_tab_size
      if int_array_get('tab_size_count', tab_size)
        AddLine(Format(tab_size:10, int_array_get('tab_size_count', tab_size):10))
      endif
    endfor

    last_log_line = CurrLine()

    if first_log_line < last_log_line
      PushBlock()
      MarkLine(first_log_line, last_log_line)
      Sort()
      PopBlock()
    endif

    EndLine()
    AddLine('')
    AddLine(Format('Best tab size:'; best_tab_size))
    AddLine('')
    SetUndoOn()
    BegFile()
    UpdateDisplay()
  end get_tab_size_debug_report
#endif

/*
  This proc counts each indentation increase as an occurrence of a tab size
  if the new indentation is a multiple of the increase,
  and it also counts it as an occurrence of the same tab size if such an
  indentation is continued on a consecutive line.

*/
integer proc get_tab_size()
  integer best_tab_size        = 0
  integer best_tab_size_count  = 0
  integer indentation          = 0
  integer indentation_increase = 0
  integer max_tab_size         = 0
  integer prev_indentation     = 1
  integer tab_size             = 0
  string  text  [MAXSTRINGLEN] = ''

  if NumLines()
    int_array_create('tab_size_count')
    PushLocation()
    BegFile()

    repeat
      text = GetText(1, MAXSTRINGLEN)

      if not Pos(Chr(9), text)
      or     Pos(Chr(9), text) > PosFirstNonWhite()
        indentation = PosFirstNonWhite()

        if indentation
          indentation_increase = indentation - prev_indentation

          if indentation_increase >= 0
            if indentation_increase             // > 0
              if (indentation - 1) mod indentation_increase
                // Invalid tab size: negative dummy value to ignore this line.
                indentation_increase = -1
              else
                tab_size = indentation_increase // New tab size.
              endif
            else
              // Same indentation: positive dummy value to count last tab size.
              indentation_increase = +1
            endif
          else
            tab_size = 0 // Negative increase: ignore line and restart scratch.
          endif

          if  tab_size
          and indentation_increase >= 0
            int_array_set('tab_size_count',
                          tab_size,
                          int_array_get('tab_size_count', tab_size) + 1)
          endif

          prev_indentation = indentation

          if tab_size > max_tab_size
            max_tab_size = tab_size
          endif
        endif
      endif
    until not Down()

    PopLocation()

    for tab_size = 1 to max_tab_size
      if int_array_get('tab_size_count', tab_size) > best_tab_size_count
        best_tab_size_count = int_array_get('tab_size_count', tab_size)
        best_tab_size       = tab_size
      endif
    endfor

    #ifdef GET_TAB_SIZE_DEBUG
      get_tab_size_debug_report(best_tab_size, max_tab_size)
    #endif

    int_array_delete('tab_size_count')
  endif

  return(best_tab_size)
end get_tab_size


