/*
  Tool            get_tab_size_test
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE v4 upwards, all variants
  Version         v2   16 May 2025

  This tool tests the get_tab_size() function from get_tab_size.inc on all text
  files in a folder, usually by supplying just the folder name (without
  wildcard characters), though you can supply just a single filename too.

  If you supply a folder, then subfolders are tested too.
  Default non-text files and files > 100,000 bytes are skipped.

  Notes:
    Files are tested after being fully opened.
    This unfortunately causes screen updates for each file, which slows down
    testing.
    For users of the Unicode extension, the benefit is, that for UTF-16 and
    UTF-32 files tab sizes are counted after their Unicode to Windows-1252
    conversion, and are therefore counted correctly.
    As this implies, for users without the Unicode extension get_tab_size()
    will not work correctly for UTF-16 and UTF-32 files, but because they
    cannot corectly edit UTF-16 and UTF-32 files anyway, this does not matter.
    It is however something to take into account when manually veryfying the
    tab sizes that this tool reports.



  INSTALLATION AND USAGE

    Obviously you need to install get_tab_size.inc first.

    Then copy this file to TSE's "mac" folder, and compile and execute it,
    for example by opening it there in TSE and applying the Macro Compile menu.

    Default it offers to determine the tab sizes of all text files in your TSE
    folder.

    In the generated report:
    - Column 1 can be:
        "Txt" for a text file
        "Bin" for a binary file.
      Default binary files are not listed, because they have no tab size anyway.
    - Column 2 can be:
        "Spc": The file contains no tab characters at all.
        "Tab": The file contains at least one tab character.
        "Tsv": The file consists of tab separated values.
      Tab separated value (TSV) files were a special case I was interested in.
      This type of file turns out to be very rare, so I decided to ignore it.

    Note:
      For ordinary distributed macros I would copy/paste the source of a .inc
      file into the macro, but for testing it makes more sense
      to let get_tab_size_test.s use an include-statement for get_tab_size.inc.




  HISTORY

  v1    24 Aug 2021
    Initial release of GetTabSize_Test.

  v2    16 May 2025.
    Renamed to get_tab_size_test.

*/




#Include ['get_tab_size.inc']





//  Start of copied is_text_file.inc ...

/*
  Include-file    is_text_file.inc
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE v4 upwards, all variants
  Version         v1   16 May 2025


  For macro programmers this include-file implements the function

    integer is_text_file(string file_name)

  It checks the named file on disk.

  It returns TRUE if the named file is a text file,
  and FALSE if it is not or if the file does not exist on disk.

  The function checks whether the named file overwhelmingly consists of
  human-readable characters. That is not a hard criterium given the lot of
  semi-readable texts out there, but the results match my expectations.


  USAGE

  - Either include is_text_file.inc in your macro with an #include statementment
    like
      #include ["is_text_file.inc"]
    or copy/paste is_text_file.inc's source code as a whole into your macro.

    Because I distribute macros, I use the copy/paste method, which avoids that
    users need to do dependency management on versions of macros and .inc files.

*/

integer is_text_file_id = 0

integer proc is_text_file(string file_name)
  integer bytes_read          = 0
  string  data [MAXSTRINGLEN] = ''
  integer handle              = 0
  integer i                   = 0
  integer org_id              = GetBufferId()
  integer result              = FALSE
  integer text_chars          = 0

  if FileExists(file_name)
    if is_text_file_id
      GotoBufferId(is_text_file_id)
      EmptyBuffer()
    else
      is_text_file_id = CreateTempBuffer()
      BinaryMode(MAXLINELEN)
    endif

    handle = fOpen(file_name)

    if handle <> -1
      repeat
        bytes_read = fRead(handle, data, MAXSTRINGLEN)
      until     bytes_read <= 0
         or not InsertText(data, _INSERT_)
         or     CurrLineLen() >= MAXLINELEN - MAXSTRINGLEN
      fClose(handle)
    endif

    for i = 1 to CurrLineLen()
      // The 0 is to accept Western UTF-16 and UTF-32 files as text.
      if (CurrChar(i) in 0, 9, 10, 12, 13, 32 .. 255)
        text_chars = text_chars + 1
      endif
    endfor

    if text_chars
      if (text_chars * 100) / CurrLineLen() >= 99   // %
        result = TRUE
      endif
    endif

    EmptyBuffer()
    GotoBufferId(org_id)
  endif

  return(result)
end is_text_file

//  End of copied is_text_file.inc.





// Global constants
#ifdef LINUX
  string SLASH [1] = '/'
#else
  string SLASH [1] = '\'
#endif

// Global variables
integer log_id  = 0


integer proc get_tabs_in_line()
  integer result = 0

  if lFind(Chr(9), 'cg')
    result = 1

    while lFind(Chr(9), 'c+')
      result = result + 1
    endwhile
  endif

  return(result)
end get_tabs_in_line


integer msg_opportunities = 0
integer msg_prev_seconds  = 0


proc get_file_tab_size(string file_name)
  string  file_format         [3] = 'Bin'
  integer file_id                 = 0
  integer lines_checked_for_tsv   = 0
  integer ok                      = TRUE
  integer old_RemoveTrailingWhite = Set(RemoveTrailingWhite, OFF)
  integer org_id                  = GetBufferId()
  integer tab_size                = 0
  string  tabs_indicator      [3] = '   '
  integer tabs_in_line            = 0
  integer tabs_per_line           = -1
  integer tsv_file                = TRUE

  msg_opportunities = msg_opportunities + 1
  if msg_opportunities mod 1000 == 0
    if GetTime() / 100 <> msg_prev_seconds
      msg_prev_seconds = GetTime() / 100
      Message(msg_opportunities, ':'; file_name)
      KeyPressed()
    endif
  endif

  if is_text_file(file_name)
    file_format = 'Txt'
    file_id     = EditThisFile(file_name, _DONT_PROMPT_)

    if file_id

      tabs_indicator = iif(lFind('\d009', 'gx'), 'Tab', 'Spc')

      if tabs_indicator == 'Tab'
        // Detect a tab separated value (TSV) file as a special case.
        BegFile()
        repeat
          if CurrLineLen()
          or lFind(Chr(9), 'cg')
            lines_checked_for_tsv = lines_checked_for_tsv + 1
            tabs_in_line          = get_tabs_in_line()

            if tabs_per_line == -1
              tabs_per_line = tabs_in_line
            elseif tabs_in_line <> tabs_per_line
              tsv_file      = FALSE
            endif
          endif
        until not tsv_file
           or not Down()
        if  tsv_file
        and lines_checked_for_tsv >= 2
          tabs_indicator = 'Tsv'
        endif
      endif

      tab_size = get_tab_size()

      GotoBufferId(org_id)
      AbandonFile(file_id)
    else
      ok = FALSE
    endif
  else
    ok = FALSE
    //  AddLine(Format(file_format; tabs_indicator;
    //                 tab_size:3;
    //                 file_name), log_id)
  endif

  if ok
    AddLine(Format(file_format; tabs_indicator;
                   tab_size:3;
                   file_name), log_id)
  endif

  Set(RemoveTrailingWhite, old_RemoveTrailingWhite)
end get_file_tab_size


proc get_dir_tab_sizes(string dir)
  integer handle = FindFirstFile(dir + SLASH + '*', -1)

  KeyPressed()
  Message(dir)
  KeyPressed()

  if handle <> -1
    repeat
      if SubStr(FFName(), 1, 1) <> '.'  // Ignores ., .., .git, etcetera.
        if FFAttribute() & _DIRECTORY_
          get_dir_tab_sizes(dir + SLASH + FFName())
        else
          if FFSize() in 0 .. 100000
            get_file_tab_size(dir + SLASH + FFName())
          endif
        endif
      endif
    until not FindNextFile(handle, -1)

    FindFileClose(handle)
  endif
end get_dir_tab_sizes


string proc unquote(string s)
  string result [MAXSTRINGLEN] = s

  if (result[1: 1] in '"', "'")
  and result[1: 1] == result[1: Length(result)]
    result = result[2: Length(result) - 2]
  endif

  return(result)
end unquote


proc Main()
  string  dir_or_file [MAXSTRINGLEN] = ''
  string  macro_name  [MAXSTRINGLEN] = SplitPath(CurrMacroFilename(), _NAME_)
  integer old_Beep                   = 0
  integer old_MsgLevel               = 0

  dir_or_file = GetHistoryStr(GetFreeHistory(macro_name + ':dir_or_file'), 1)

  if dir_or_file == ''
    dir_or_file = QuotePath(RemoveTrailingSlash(LoadDir()))
  endif

  if Ask('Get the tab sizes for which file, or for all files in which folder:',
         dir_or_file,
         GetFreeHistory(macro_name + ':dir_or_file'))

    BufferVideo()
    old_Beep     = Set(Beep    ,  OFF  )
    old_MsgLevel = Set(MsgLevel, _NONE_)

    dir_or_file = RemoveTrailingSlash(unquote(Trim(dir_or_file)))
    log_id = NewFile()

    if FileExists(dir_or_file) & _DIRECTORY_
      get_dir_tab_sizes(dir_or_file)
    else
      get_file_tab_size(dir_or_file)
    endif

    GotoBufferId(log_id)
    BegFile()

    Set(Beep    , old_Beep    )
    Set(MsgLevel, old_MsgLevel)
    UnBufferVideo()

  endif

  PurgeMacro(SplitPath(CurrMacroFilename(), _NAME_))
end Main

